/**************************************************************************

        @file bmscanner.d

        Copyright (c) 2005 Derek Parnell

        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.

        Permission is hereby granted to anyone to use this software for any
        purpose, including commercial applications, and to alter it and/or
        redistribute it freely, subject to the following restrictions:

        1. The origin of this software must not be misrepresented; you must
           not claim that you wrote the original software. If you use this
           software in a product, an acknowledgment within documentation of
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2005
        @author         Derek Parnell, Melbourne, Australia.
        @acknowledgment Thanks to Robert Boyer and J. Strother Moore
                        for their string finding algorithm.
                        (Communications of the Association for Computing
                         Machinery, 20(10), 1977, pp. 762-772.)

***************************************************************************
This is an adaption of the famous Boyer-Moore Fast String Searching Algorithm
written in the D programming language ( http://www.digitalmars.com/d/ )

This module contains a template for instantiating functions for any
dynamic array type.

The function returns an int. If this is returned as -1 then the Find
array does not exist in the Container array, otherwise it returns the
offset (index) into the container array.

The function receives two mandatory arguments and one optional argument.
The first argument is the array to search. It may or may not contain the
second argument. Both the first and second argument must be dynamic
arrays of the same type. The optional third argument is a starting offset
into the first argument. By default it is zero (0).

Example:
    alias BMScan!(char) bmfind;
    int pos1;
    int pos2;

    // Locate the first and second occurances.
    pos1 = bmfind(TextLine, Fragment);
    if (pos1 != -1)
        pos2 = bmfind(TextLine, Fragment, pos1 + 1);
    else
        pos2 = -1;

*/
module util.bmscanner;
private import util.bmscanner_bn;

template BMScan(T) {
int BMScan(T[] pContainer, T[] pFind, uint pStartPos=0)
{
    int[T] lShift; // Shift table.
    int lCP;    // Index into container array
    int lCE;    // Current high-water mark in containter array
    int lFP;    // Index into find array

    // Do length checks first.
    if (pFind.length == 0)
        return -1;
    if (pStartPos > pContainer.length)
        return -1;
    if (pStartPos < 0)
        return -1;
    if (pStartPos > 0)
        pContainer = pContainer[pStartPos .. $];
    if (pContainer.length == 0)
        return -1;
    if (pFind.length > pContainer.length)
        return -1;

    // If find string is only one element long, do a brute force scan.
    if (pFind.length == 1)
    {
        T pElem = pFind[0];
        foreach(int i, T lElem; pContainer[pStartPos .. $])
        {
            if ( lElem == pElem)
                return i + pStartPos;
        }
        return -1;
    }

    // Prepare shift table.
    foreach(int i, T c; pFind)
    {
        lShift[c] = pFind.length - i - 1;
    }

    // Start scan at rightmost position.
    lCP = pFind.length - 1;
    while(lCP < pContainer.length)
    {
        lFP = pFind.length - 1;
        assert(lCP == 0 || lCP > lCE);
        lCE = lCP;  // Remember where I'm up to so far.

        if (pContainer[lCP] != pFind[lFP])
        {   // Found a mismatch, so realign the arrays.
            if ((pContainer[lCP] in lShift) == null)
            {
                // This element is not anywhere in the Find array
                // so we can skip over it to align up just beyond it.
                lCP += pFind.length;
            }
            else
            {
                // This element is somewhere in the Find array
                // so if we haven't checked all Find elements we
                // can slide up to align with the rightmost
                // position where the element occurs in the Find
                // array. However, if we have checked the entire
                // Find array, then it isn't in the Container array.
                if (lFP == 0)
                    return -1;
                lCP += lShift[pContainer[lCP]];
            }
        }
        else
        {   // Found a match, so scan right-to-left...
            // unless this we are at the first Find array element
            // in which case we have found it.
            if (lFP == 0)
               return lCP + pStartPos;

            // Keep scanning until we run out of elements in the Find array.
            while(lFP > 0)
            {
                // Point to next pair to check.
                lFP--;
                lCP--;

                if (pContainer[lCP] != pFind[lFP])
                {   // Found a mismatch, so realign the arrays.
                    if ((pContainer[lCP] in lShift) == null)
                    {
                        // This element is not anywhere in the Find array
                        // so we can skip over it to align up just beyond it.
                        lCP += pFind.length;
                    }
                    else
                    {
                        // This element is somewhere in the Find array
                        // so if we haven't checked all Find elements we
                        // can realign the match so far beyond the current
                        // 'block'.
                        // However, if we have checked the entire
                        // Find array, then it isn't in the Container array.
                        if (lFP == 0)
                            return -1;
                        lCP = lCE + (pFind.length - lFP - 1);
                    }
                    // Stop element-by-element scanning and resume
                    // block-end checking.
                    break;
                }

                if (lFP == 0)
                    // We are at the begining of the Find array so
                    // that means we found a complete match!
                    return lCP + pStartPos;
            }
        }

    }
    // If we get here, then there is no match.
    return -1;

}
private import std.stdio;
int BMScanElement(T[] pContainer, T pFind, uint pStartPos=0)
{
    foreach(int i, T lElem; pContainer[pStartPos .. $])
    {
        if ( lElem == pFind)
        {
            return i + pStartPos;
        }
    }
    return -1;
}
}

unittest{

    alias BMScan!(char).BMScan utfind;
    alias BMScan!(char).BMScanElement utfind_elem;


    assert( 7 == utfind_elem("a simple example to demonstrate Boyer-Moore", 'e'));
    assert( 9 == utfind_elem("a simple example to demonstrate Boyer-Moore", 'e', 8));
    assert(-1 == utfind_elem("a simple example to demonstrate Boyer-Moore", 'z'));
    assert( 9 == utfind("a simple example to demonstrate Boyer-Moore", "example"));
    assert(10 == utfind("a simple example to demonstrate Boyer-Moore", "xample"));
    assert( 4 == utfind("a simple example to demonstrate Boyer-Moore", "mple"));
    assert(12 == utfind("a simple example to demonstrate Boyer-Moore", "mple", 5));
    assert( 6 == utfind("a simple example to demonstrate Boyer-Moore", "le"));
    assert( 3 == utfind("a simple example to demonstrate Boyer-Moore", "i"));
    assert(-1 == utfind("a simple example to demonstrate Boyer-Moore", "xy"));
    assert(41 == utfind("a simple example to demonstrate Boyer-Moore", "re"));
    assert(-1 == utfind("a simple example to demonstrate Boyer-Moore", ""));
    assert(-1 == utfind("", ""));
    assert( 0 == utfind("a", "a"));
    assert(-1 == utfind("a", "b"));
    assert(-1 == utfind("a", "aa"));
    assert(-1 == utfind("", "xyzzy"));
    assert(-1 == utfind("xy", "xyzzy"));
    assert( 0 == utfind("xyzzy", "xyzzy"));
    assert( 1 == utfind("xyzzyz", "yz"));
    assert(12 == utfind("a simple example to demonstrate Boyer-Moore", "mple to demonstrate Boye"));
    assert( 7 == utfind("abbbbcbbbbd", "bbbd"));
    assert( 7 == utfind("bbbbbbbbbbd", "bbbd"));
    assert(-1 == utfind("bbbbbbbbbbd", "dbbb"));
    assert(-1 == utfind("abababababa", "bb"));
    assert( 1 == utfind("abababababa", "ba"));
    assert( 1 == utfind("abababababa", "ba", 1));
    assert( 7 == utfind("abababababa", "ba", 6));
    assert( 3 == utfind("abccddefghijkl", "cdd"));
    assert( 2 == utfind("abccddefghijkl", "ccdd"));
    assert(-1 == utfind("abccddefghijkl", "cccdd"));
    assert( 4 == utfind("abccddefghijkl", "dde"));
    assert(17 == utfind("abcccddeeffghijkldcde", "dcde"));
}