// Author: Andrew Fedoniouk
// Date: 22 April 2005


module util.custsort;
// "smart" sort implementation, works 1.3 times faster than D builtin sort

template sort(T)
{
  T[] sort(T[] arr, int function(in T l, in T r) cmp)
  {
      const int quick_sort_threshold = 9;

      struct tuple
      {
          int low;
          int high;
      }

      if(arr.length < 2) return arr;

      void swap_if_less( inout T t1, inout T t2 )
      {
        if( cmp(t1, t2) < 0 ) { T t = t1; t1 = t2; t2 = t; }
      }
      void swap( inout T t1, inout T t2 )
      {
        T t = t1; t1 = t2; t2 = t;
      }

      tuple[] stack;
      stack.length = 0;
      int   top = 0;
      int   limit = arr.length;
      int   base = 0;

      for(;;)
      {
          int len = limit - base;

          int i;
          int j;
          int pivot;

          if(len > quick_sort_threshold)
          {
              // we use base + len/2 as the pivot
              pivot = base + len / 2;
              swap(arr[base], arr[pivot]);

              i = base + 1;
              j = limit - 1;

              // now ensure that *i <= *base <= *j
              swap_if_less(arr[j], arr[i]);
              swap_if_less(arr[base],arr[i]);
              swap_if_less(arr[j],arr[base]);

              for(;;)
              {
                  do i++; while( cmp(arr[i],arr[base]) < 0 );
                  do j--; while( cmp(arr[base],arr[j]) < 0 );
                  if( i > j )
                  {
                      break;
                  }
                  swap(arr[i], arr[j]);
              }

              swap(arr[base], arr[j]);

              // now, push the largest sub-array
              if(top >= stack.length)
                 // need more stack space so expand it.
                 stack.length = stack.length + 16;
              if(j - base > limit - i)
              {
                  stack[top].low = base;
                  stack[top].high = j;
                  base   = i;
              }
              else
              {
                  stack[top].low = i;
                  stack[top].high = limit;
                  limit  = j;
              }
              top++;
          }
          else
          {
              // the sub-array is small, perform insertion sort
              j = base;
              i = j + 1;

              for(; i < limit; j = i, i++)
              {
                  for(; cmp(arr[j + 1],arr[j]) < 0; j--)
                  {
                      swap(arr[j + 1],arr[j]);
                      if(j == base)
                          break;
                  }
              }
              if(top > 0)
              {
                  top--;
                  base = stack[top].low;
                  limit = stack[top].high;
              }
              else
                  break;
          }
      }

      return arr;
  }
}

alias sort!(char)   csort;
alias sort!(wchar)  csort;
alias sort!(dchar)  csort;
alias sort!(byte)   csort;
alias sort!(short)  csort;
alias sort!(int)    csort;
alias sort!(long)   csort;
alias sort!(ubyte)  csort;
alias sort!(ushort) csort;
alias sort!(uint)   csort;
alias sort!(ulong)  csort;
alias sort!(float)  csort;
alias sort!(double) csort;
alias sort!(real)   csort;

/* Usage:
    char[] a;
    a.csort(&mycomparator);
*/

  private int less(char a, char b)
  {
    // Upper case is always sorted after than lowercase.
    int A = a; if(a >= 'A' && a <='Z') A <<= 8;
    int B = b; if(b >= 'A' && b <='Z') B <<= 8;
    return A - B;
  }

unittest{

  char[] arr;
  arr = "KlMnOpQrStUvWxYzAbCdEfGhIj".dup;
  assert(arr.csort(&less) == "bdfhjlnprtvxzACEGIKMOQSUWY");
  arr = "Kl".dup;
  assert(arr.csort(&less) == "lK");
  arr = "l".dup;
  assert(arr.csort(&less) == "l");
  arr = "".dup;
  assert(arr.csort(&less) == "");
  arr = "KlMnOpQrStUvWxYzAbCdEfGhIjKlMnOpQrStUvWxYzAbCdEfGhIjKlMnOpQrStUvWxYzAbCdEfGhIjKlMnOpQrStUvWxYzAbCdEfGhIj".dup;
  assert(arr.csort(&less) == "bbbbddddffffhhhhjjjjllllnnnnpppprrrrttttvvvvxxxxzzzzAAAACCCCEEEEGGGGIIIIKKKKMMMMOOOOQQQQSSSSUUUUWWWWYYYY");
  arr = "9876543210zyxwvutsrqponmlkjihgfedcba".dup;
  assert(arr.csort(&less) == "0123456789abcdefghijklmnopqrstuvwxyz");

}
